#lang racket

(provide
  variable
  variable?
  variable-identifier
  variable-line-number
  variable-user
  field-instance
  field-instance?
  field-instance-identifier
  field-instance-path
  field-instance-line-number
  field-instance->identifier
  get-new-project-modifier-struct
  merge-two-modifier-structures
  modifies-left?
  modifies-right?
  modifies-start?
  modifies-stop?
  region-chr
  region-left-end
  region-right-end
  region-strand
  modified-left
  modified-right
  create-modifier
  project-modifier-struct-left-modifier
  project-modifier-struct-right-modifier
  project-modifier-struct-start-modifier
  project-modifier-struct-stop-modifier)


;this data structure is used to store the 
;identifier of a variable along with the line number
;in which it appears.
;This is useful for error printing
(struct variable
	(identifier
	 line-number
	 user)) ;user who owns the dataset

;struct to store any instance of field reference
(struct field-instance
	(identifier
	 line-number
	 path)) ; the path to the field, could be either full or partial

;given a field-instance, build a string a.b.name
(define (field-instance->identifier field)
  (let ((path (field-instance-path field)))  
    (foldr 
      (λ (x y) (string-append x "." y)) 
      (field-instance-identifier field)
      (if (list? path) path (list)))))


;this struct stores the project modifier
(struct project-modifier-struct
	([left-modifier  #:mutable]	;a symbol if empty, a string is a condition is set
	 [right-modifier	  #:mutable]	; ''
	 [start-modifier  #:mutable]	; ''
	 [stop-modifier	  #:mutable]	; ''
	 [value-modifiers #:mutable]))	;a list (eventually empty) of value modifiers

(define (get-new-project-modifier-struct left right start stop mod-list)
  (project-modifier-struct left  right start stop mod-list))

;merge two modifier data structure checking if they are compatible
(define (merge-two-modifier-structures st1 st2)
  (letrec
    ((left-st1 (not (symbol? (project-modifier-struct-left-modifier st1))))
     (right-st1 (not (symbol? (project-modifier-struct-right-modifier st1))))
     (start-st1 (not (symbol? (project-modifier-struct-start-modifier st1))))
     (stop-st1 (not (symbol? (project-modifier-struct-stop-modifier st1))))
     (left-st2 (not (symbol? (project-modifier-struct-left-modifier st2))))
     (right-st2 (not (symbol? (project-modifier-struct-right-modifier st2))))
     (start-st2 (not (symbol? (project-modifier-struct-start-modifier st2))))
     (stop-st2 (not (symbol? (project-modifier-struct-stop-modifier st2))))
     (fixed-st1 (and left-st1 right-st1))
     (fixed-st2 (and left-st2 right-st2))
     (relative-st1 (and start-st1 stop-st1))
     (relative-st2 (and start-st2 stop-st2)))
    (cond
      ((or (and fixed-st1 relative-st2) (and relative-st1 fixed-st2))
	   (error "ERROR: region modifier are not consistent; cannot combine relative modifier (start, stop) with fixed modifier (left-right)"))
      ((and left-st1 left-st2) (error "ERROR: region modifier on 'left' is repeated more than once"))
      ((and right-st1 right-st2) (error "ERROR: region modifier on 'right' is repeated more than once"))
      ((and start-st1 start-st2) (error "ERROR: region modifier on 'start' is repeated more than once"))
      ((and stop-st1 stop-st2) (error "ERROR: region modifier on 'stop' is repeated more than once"))
      (else (let
	      ((new-left (cond (left-st1 (project-modifier-struct-left-modifier st1)) (left-st2 (project-modifier-struct-left-modifier st2)) (else 'empty)))
	       (new-right (cond (right-st1 (project-modifier-struct-right-modifier st1)) (right-st2 (project-modifier-struct-right-modifier st2)) (else 'empty)))
	       (new-start (cond (start-st1 (project-modifier-struct-start-modifier st1)) (start-st2 (project-modifier-struct-start-modifier st2)) (else 'empty)))
	       (new-stop (cond (right-st1 (project-modifier-struct-stop-modifier st1)) (stop-st2 (project-modifier-struct-stop-modifier st2)) (else 'empty)))
	       (new-list (append (project-modifier-struct-value-modifiers st1) (project-modifier-struct-value-modifiers st2))))
	      (get-new-project-modifier-struct new-left new-right new-start new-stop new-list))))))


(define (modifies-left? st)  (not (symbol? (project-modifier-struct-left-modifier st))))
(define (modifies-right? st)  (not (symbol? (project-modifier-struct-right-modifier st))))
(define (modifies-start? st)  (not (symbol? (project-modifier-struct-start-modifier st))))
(define (modifies-stop? st)  (not (symbol? (project-modifier-struct-stop-modifier st))))


(define (modified-left st) (if (modifies-left? st)  (project-modifier-struct-left-modifier st) "$1.$1"))
(define (modified-right st) (if (modifies-right? st) (project-modifier-struct-right-modifier st) "$1.$2"))

;given: starting OP offset
;generate the corresponding arguments for the helper function
; (starting, offset)
(define (create-modifier starting op offset)
  (let
    ((os (if (equal? "+" op) offset (- offset))))
    (list starting os)))

;some useful constants
(define region-chr "regions.$0")
(define region-left-end "region.$1")
(define region-right-end "region.$2")
(define region-strand "regions.$3")
