#lang racket

(require 
  "data-structures.rkt"
  (prefix-in schemata-manager: "schemata-manager.rkt")
  (prefix-in xml-utils: "xml-utils.rkt")
  "configurations.rkt"
  "read-command-line.rkt"
  racket/list) 

(provide
  emit-load-metadata-code
  emit-select-code
  emit-project-code
  emit-map-code
  emit-materialize-code
  emit-join-code
  emit-union-code
  emit-region-intersection-code
  emit-difference-code
  emit-aggregate-code
  emit-select-all-code
  emit-order-code
  emit-pre-parsing-code
  emit-post-parsing-code
  )

;every emit-OPERATOR- function returns a list
;(list preamble-code pig-code)

;this function emit code for registring the java package
; and for loading the metada files
(define (emit-pre-parsing-code scm)
  (let
    ((pig "")
     (preamble ""))
    (begin
      (set! pig (string-append pig "\nregister " library-path ";\n"))
      (map
	(λ (x)
	   (let
	     ((name (first x))
		    (user (second x)))
	     (set! pig (string-append pig (second (emit-load-metadata-code user name)) "\n"))))
	(scm 'get-list))
      (list preamble pig))))

;In the Pig Lating implementation, this does nothing
(define (emit-post-parsing-code scm)
  (list "" ""))

;utility function to inform the user that a particular
;schema is not definied
(define (not-existing-schema-error var)
  (letrec
    ((var-name (variable-identifier var))
     (var-line (variable-line-number var))
     (var-user (variable-user var)))
    (error
      (string-append 
	"variable " var-name " at line " (number->string var-line) " has not been defined " 
	(if (string? var-user)
	  (string-append "for user " var-user)
	  "")))))

;utility function to inform user about a schema that
;has multiple definitions
(define (multiple-schema-definition-error scm var)
  (letrec 
    ((var-name (variable-identifier var))
     (var-line (variable-line-number var))
     (usr-list (scm 'variable-name->user-list var-name))
     (list-msg (foldr (λ (x y) (string-append "\t" x "\n" y)) "" usr-list)))
    (error
      (string-append "variable name " var-name " at line " (number->string var-line) " is ambiguos. " 
		     "More than one user define it, please specify one. "
		     "Available options are: \n" list-msg))))

;warn the user if he is trying to perform an union on dataset with different schema
(define (different-schema-error var1 var2)
  (let
    ((var1-name (variable-identifier var1))
     (var1-line (variable-line-number var1))
     (var2-name (variable-identifier var2))
     (var2-line (variable-line-number var2)))
    (error (string-append
	     "variable "
	     var1-name
	     " at line "
	     (number->string var1-line)
	     " does not have the same schema of variable "
	     var2-name
	     " at line "
	     (number->string var2-line)))))

;utilify function to inform user that a 
;field is not defined for a certain dataset
(define (not-existing-field-error field-name field-number var-name)
  (error
    (string-append "field '" field-name "' at line " (number->string field-number)
		   " is not defined for variable '" var-name "'")))

;print error when a field has multiple definition in the schema
(define (multiple-field-definition field-name field-number var-name)
  (error
    (string-append "field '" field-name "' at line " (number->string field-number)
		   "has multiple definition. "
		   "Please, specify a path to select one of them")))

;utility function toinform the user about a redefinition
;of a variable
(define (variable-redefinition-error var)
  (let
    ((var-name (variable-identifier var))
     (var-line (variable-line-number var)))
    (error
      (string-append "variable " var-name " has been already defined. "
		     "Cannot be redefined at line " (number->string var-line)))))

;translates a PROJECT condition in a PigLatin predicate 
(define (translate-expression scm x var)
  (cond
   [(equal? x 'TRUE) "true"]
    [(and (member (car x) (list "==" "<" ">" "<=" ">=" "!=")) (member (second x) (list 'chr 'left 'right 'strand))) 
     (let
       ((operator (car x))
	(left-operand (cond
			((equal? (second x) 'chr) "$1.$0")
			((equal? (second x) 'left) "$1.$1")
			((equal? (second x) 'right) "$1.$2")
			((equal? (second x) 'strand) "$1.$3")))
	(right-operand (cond
			 ((member (second x) (list 'chr 'strand)) (string-append "'" (third x) "'"))
			 (else (number->string (third x))))))
       (string-append left-operand operator right-operand))]
    ;if the root of the expression 'x' is an operator (and therefore x is a check...)
    [(member (car x) (list "==" "<" ">" "<=" ">=" "!=")) 
     (letrec
       ((var-name (variable-identifier var))
	(var-owner (variable-user var))
	(op (first x))
	(right-operand (cond 
			 ((string? (third x)) (third x))
			 ((number? (third x)) (number->string (third x)))
			 ((field-instance? (third x)) (third x))))
	(right-field-name (when (field-instance? right-operand) (field-instance-identifier right-operand)))
	(right-field-path (when (field-instance? right-operand) (field-instance-path right-operand)))
	(right-field-line (when (field-instance? right-operand) (field-instance-line-number right-operand)))
	(right-value (if 
		       (field-instance? right-operand)
		       (let
			 ((pos-list (scm 'position var-name var-owner right-field-name right-field-path)))
			 (string-append "value.$" 
					(number->string 
					  (cond
					    ((equal? (length pos-list) 0) (not-existing-field-error right-field-name right-field-line var-name))
					    ((> (length pos-list) 1) (multiple-field-definition right-field-name right-field-line var-name))
					    (else (car pos-list))))))
		       right-operand)) ;it is a string
	(left-operand (second x)) ;the left operand is always a field
	(left-field-name (field-instance-identifier left-operand))
	(left-field-path (field-instance-path left-operand))
	(left-field-line (field-instance-line-number left-operand))
	(left-position 
	  (let ((pos-list (scm 'position var-name var-owner left-field-name left-field-path)))
	    (cond
	      ((equal? (length pos-list) 0) (not-existing-field-error left-field-name left-field-line var-name))
	      ((> (length pos-list) 1) (multiple-field-definition left-field-name left-field-line var-name))
	      (else (car pos-list))))))
       (string-append "value.$" (number->string left-position) " " op " " right-value))]
    [(list? (car x)) (string-append "(" (translate-expression scm (car x) var) ")" )]
    [(equal? (car x) "or") (string-append (translate-expression scm (second x) var) " or "(translate-expression scm (third  x) var))]
    [(equal? (car x) "and") (string-append (translate-expression scm (second x) var) " and "(translate-expression scm (third x) var))]))


(define (identifier-of var)
  (cond
    ((variable? var) 
     (let ((user (if (string? (variable-user var)) (variable-user var) "unknown")))
       (string-append user "_" (variable-identifier var))))
    ((string? var) var)
    (else (error "internal error (in parser-utils:identifier-of)."))))


(define (get-meta-name var)
  (string-append (identifier-of var) "_meta"))

(define (get-exp-name var)
  (string-append (identifier-of var) "_exp"))

(define (get-ids-name var)
  (string-append (identifier-of var) "_ids"))


(define (join-stm left-var left-field right-var right-field . type)
  (string-append 
    "join " left-var " by " left-field " , " right-var " by " right-field 
    (if (not (null? type)) (string-append " using '" (car type) "' ") " ") 
    (get-parallelism)))

(define (get-parallelism)
  (if (> parallelism-level 1)
    (string-append "parallel " (number->string parallelism-level))
    ""))

(define (comma-separated-list x . xs)
  (letrec
    ([l (cons x xs)]
     [f (lambda (a)
	  (if (null? (cdr a))
	    (car a)
	    (string-append (car a) ", " (f (cdr a)))))])
    (f l)))

(define-syntax-rule (code-append code stm)
		    (set! code (string-append code stm ";\n")))

(define-syntax-rule (assignment-stm left right)
		    (string-append left " = " right))


(define-syntax-rule (foreach-stm alias field)
		    (string-append "foreach " alias " generate " field))

(define-syntax-rule (nested-foreach-stm alias inner-assignment-list to-be-generated)
		    (string-append
		      "foreach " alias "{" (string-join inner-assignment-list ";") "; generate " to-be-generated ";}")) 

(define-syntax-rule (par code)
		    (string-append "( " code " )"))

(define-syntax-rule (as-stm value alias)
		    (string-append value " as " alias))

(define-syntax-rule (define-meta-filter var condition)
		    (let
		      ((var-name (identifier-of var))
		       (constructor "GenometricPig.MetaFilter"))
		      (string-append "define " var-name "_filter " constructor "('" condition "')")))

(define-syntax-rule (define-region-modifier var mods)
		    (let
		      ((var-name (identifier-of var))
		       (constructor "GenometricPig.RegionModifier")
		       (first-start (first (first mods)))
		       (first-op (string-append (second (first mods))  (number->string (third (first mods)))))
		       (second-start (first (second mods)))
		       (second-op (string-append (second (second mods)) (number->string (third (second mods))))))
		      (string-append "define " var-name "_modifier " constructor "("
				     "'" first-start "','" first-op "','" second-start "','" second-op "')")))

(define-syntax-rule (define-difference-function var type)
		    (let
		      ((var-name (identifier-of var))
		       (constructor "GenometricPig.Difference"))
		      (string-append "define " var-name "_difference " constructor "('" type "')")))

(define-syntax-rule (get-difference-function var)
		    (let
		      ((var-name (identifier-of var)))
		      (string-append var-name "_difference")))

(define-syntax-rule (get-meta-group-name var)
		    (string-append (identifier-of var) "_meta_group"))

(define-syntax-rule (get-filter-function-name var)
		    (string-append (identifier-of var) "_filter"))

(define-syntax-rule (filter-stm alias expression)
		    (string-append "filter " alias " by " expression))

(define-syntax-rule (filter-field-not-null alias field)
		    (string-append "filter " alias " by " field " is not null"))

(define-syntax-rule (apply-function fname fpar)
		    (string-append fname "(" fpar ")"))

(define-syntax-rule (get-:: var field)
		    (string-append var "::" field)) 

(define-syntax-rule (group-stm alias expression)
		    (string-append "group " alias " by " expression " " (get-parallelism)))

;the argument is a list of string, e.g ("COUNT", "AVG", "1")
;type is a string, either "stranded" or "not-stranded"
(define-syntax-rule (define-map-function var argument type)
		    (let
		      ((var-name (identifier-of var))
		       (constructor "GenometricPig.Map"))
		      (string-append "define " var-name "_mapper " constructor "('" 
				     (string-append
				       (string-join argument " ")
				       " " type)
				     "')"))) 

(define-syntax-rule (define-order-function var ordering top)
  (letrec
      ((var-name (variable-identifier var))
       (constructor "GenometricPig.OrderBy")
       (new-att-name (string-append "'_order'"))
       (ord-option (string-append "'"
                    (string-join 
                    (map 
                     (lambda (x) 
                       (string-append (symbol->string (first x)) " " (field-instance-identifier (second x)))) 
                     ordering) " ")
                    "'"))
       (optional-top (if (null? top) "" (string-append " , '" (symbol->string (first top)) " " (number->string (second top)) "'"))))
    (string-append "define " (identifier-of var) "_order " constructor "("
                   ord-option ", " new-att-name optional-top ")")))

(define-syntax-rule (get-order-function var)
  (string-append (identifier-of var) "_order"))

(define-syntax-rule (get-exp-group-name var)
		    (string-append (identifier-of var) "_exp_group"))

(define-syntax-rule (accessor var field)
		    (string-append var "." field))

(define-syntax-rule (get-join-name var1 var2)
		    (string-append (identifier-of var1) "_" (identifier-of var2) "_join"))

(define-syntax-rule (get-cross-name var1 var2)
		    (string-append (identifier-of var1) "_" (identifier-of var2) "_cross"))

(define-syntax-rule (get-map-name var1 var2)
		    (string-append (identifier-of var1) "_" (identifier-of var2) "_map"))

(define-syntax-rule (get-mapper-function-name var)
		    (string-append (identifier-of var) "_mapper"))

(define-syntax-rule (distinct-stm alias)
		    (string-append "distinct ( " alias " )"))

(define-syntax-rule (store-stm var)
		    (string-append
		      (string-append "store " (get-meta-name (identifier-of var)) " into '" materialize-directory (variable-identifier var) "_meta';\n")
		      (string-append "store " (get-exp-name (identifier-of var)) " into '" materialize-directory (variable-identifier var) "_exp';\n")))

(define-syntax-rule (get-exp-cross-name var1 var2)
		    (string-append (identifier-of var1) "_" (identifier-of var2) "_exp_cross"))

(define-syntax-rule (get-meta-cross-name var1 var2)
		    (string-append (identifier-of var1) "_" (identifier-of var2) "_meta_cross"))

(define-syntax-rule (cross-stm left-var right-var)
		    (string-append
		      "cross " left-var " , " right-var " " (get-parallelism))) 

(define-syntax-rule (get-meta-filter-function-name var)
		    (string-append (identifier-of var) "_meta_filter"))

(define-syntax-rule (get-flattened-name var)
		    (string-append (identifier-of var) "_flattened"))

(define-syntax-rule (union-stm var1 var2)
		    (string-append "union " (identifier-of var1) " , " (identifier-of var2)))

(define-syntax-rule (get-joiner-function-name var)
		    (string-append (identifier-of var) "_joiner"))

(define-syntax-rule (get-meta-name-filter var)
		    (string-append (get-meta-name (identifier-of var)) "_filtered"))

(define-syntax-rule (group-all-stm alias)
		    (string-append "group " alias " ALL " (get-parallelism)))

(define-syntax-rule (get-intersection-function var)
		    (string-append (identifier-of var) "_intersection_op")) 

(define-syntax-rule (define-intersection-metadata-grouper var groups)
		    (string-append 
		      "define " 
		      (identifier-of var) "_metadata_grouper "
		      "GenometricPig.CoverMetadataGrouper("
		      (string-join
			(map
			  (λ (x)
			     (string-append "'" (field-instance-identifier x) "'"))
			  groups)
			",")
		      ")"))

(define-syntax-rule (get-cover-metadata-grouper-name var)
		    (string-append (identifier-of var) "_metadata_grouper"))

;parameters is a list of strings
(define (define-join-function var parameters)
  (let 
    ((var-name (identifier-of var))
     (constructor "GenometricPig.Join")
     (param-string
       (foldr
	 (λ (x y) (if (string? y) (string-append x "," y) x))
	 '()
	 (map (λ (x) (string-append "'" x "'")) parameters))))
    (string-append "define " var-name "_joiner "
		   constructor "(" param-string ")")))

;var: variable
;minAcc,maxAcc,minJac,maxJac: string
;type: symbol, 'cover, 'cover-flat, 'summit, 'summit-flat
;merge: boolean
(define (define-intersection-operation var type minAcc maxAcc minJac maxJac merge aggregate-function-string)
  (letrec
    ((var-name (identifier-of var))
     (constructor (string-append "GenometricPig."
				 (cond
				   ((or (equal? type 'cover) (equal? type 'cover-flat)) "Cover")
				   ((or (equal? type 'summit) (equal? type 'summit-flat)) "Summit")
				   (else (error "Internal error, invalid region intersection operation type")))))
     (is-merge (if merge "merge" "not-merge"))
     (is-flat (if (or (equal? type 'cover) (equal? type 'summit)) "not-flat" "flat"))
     (name (string-append var-name "_intersection_op")))
    (string-append "define " name " " constructor "(" 
		   (string-join
                    (append 
                     (map (λ (x) (string-append "'" x "'")) 
                          (list minAcc maxAcc minJac maxJac is-merge is-flat))
                     (list (string-append "'" aggregate-function-string "'")))
                    ",")
		   ")")))

;arg is a string
(define (define-join-meta-filter var arg)
  (letrec
    ((var-name (identifier-of var))
     (constructor "GenometricPig.JoinMetaFilter")
     (name (string-append var-name "_meta_filter")))
    (string-append "define " name " " constructor "('" arg "')")))

(define (emit-load-metadata-code user name)
  (letrec
    ((path (string-append data-repository-root "/" user "/metadata/" name ".meta" ))
     (schema "(id:long, attribute:chararray, value:chararray)")
     (left  (string-append user "_" name "_meta"))
     (preamble "")
     (pig-code (string-append left " = load '" path "' AS " schema ";")))
    (list preamble pig-code)))


(define (convert-lucene-condition condition)
  (letrec
    ((substituted (map
		    (λ (x) 
		       (cond
			 ((equal? x "*") " AND ")
			 ((equal? x "+") " OR ")
			 ((equal? x "(") "(")
			 ((equal? x ")") ")")
			 ((equal? x "=") "=")
			 ((equal? x "!=") "!=")
			 (else 
			   (list->string
			     (map
			       (λ (e)
				  (if 
				    (or
				      (equal? e #\()
				      (equal? e #\space)
				      (equal? e #\))
				      (equal? e #\,)
				      (equal? e #\;)
				      (equal? e #\:)
				      (equal? e #\.)
				      (equal? e #\&)
				      (equal? e #\%)
				      (equal? e #\\)
				      (equal? e #\/)
				      (equal? e #\|)
				      (equal? e #\!)
				      (equal? e #\^)
				      (equal? e #\+)
				      (equal? e #\-)
				      (equal? e #\[)
				      (equal? e #\])
				      (equal? e #\{)
				      (equal? e #\}))
				    #\_
				    e))
			       (string->list x))))))
		    (filter 
		      (λ (x) (not (or (equal? "[" x) (equal? "]" x)))) 
		      condition)))
     (replacer (lambda (l)
		 (if 
		   (< (length l) 3)
		   l
		   (cond
		     ((equal? (second l) "=") (cons (string-append (first l) "_" (third l)) (replacer (cdddr l))))
		     ((equal? (second l) "!=") (cons (string-append "(" (first l) "* AND "  "NOT ( " (first l) "_" (third l) " )) ") (replacer (cdddr l))))
		     (else (cons (first l) (replacer (cdr l)))))))))
    (replacer substituted)))


;;function to emit both pig and preamble when the input variable is indexed by lucene
(define (emit-lucene-select-code left right condition type user)
  (letrec
    ((left-name (identifier-of left))
     (parameter-name left-name)
     (dataset-name (string-append user "/" (variable-identifier right)))
     (preamble (string-append "-- " parameter-name "\t" dataset-name "\t" (string-join (convert-lucene-condition condition) "")"\n"))
     (loading-function "PigInputOutput.LoadRegions")
     (xml-url (string-append repository-root "/" user "/datasets/" (variable-identifier right) ".xml"))
     (schema "(id:long, region:(chr:chararray, left:int, right:int, strand:chararray), value:())")
     (pig (string-append "\n\n-- LUCENE-SELECT " left-name "\n")))
    (begin 
      (code-append 
	pig
	(string-append (get-exp-name left) " = load '$" parameter-name
		       "' using " loading-function "('" type "','" xml-url "') "
		       " AS " schema))
      (code-append 
	pig
	(assignment-stm 
	  (get-meta-name left) 
	  (foreach-stm
	    (par	   
	      (join-stm
		(string-append user "_" (variable-identifier right) "_meta") "id"
		(par (distinct-stm (foreach-stm (get-exp-name left) "id"))) "id"
		"replicated"))
	    (comma-separated-list
	      (as-stm "$0" "id")
	      (as-stm "$1" "attribute")
	      (as-stm "$2" "value")))))
      (list preamble pig))))

;emit code for selection on unindexed data
(define (emit-select-code-simple left right condition user)
  (letrec
    ((left-name (identifier-of left))
     (preamble "")
     (pig (string-append"-- SIMPLE-SELECT " left-name "\n")))
    (begin
      (code-append pig 
		   (define-meta-filter 
		     left
		     (string-join condition "#")))
      (code-append pig
		   (assignment-stm (get-meta-group-name right) (group-stm (get-meta-name right) "id")))
      (code-append pig
		   (assignment-stm 
		     (get-ids-name left) 
		     (foreach-stm 
		       (par (filter-stm (get-meta-group-name right) (apply-function (get-filter-function-name left) (get-meta-name right))))
		       (as-stm "$0" "id"))))
      (code-append pig
		   (assignment-stm
		     (get-meta-name left)
		     (foreach-stm 
		       (par (join-stm (get-meta-name right) "id" (get-ids-name left) "id" "replicated"))
		       (comma-separated-list 
			 (as-stm (get-:: (get-meta-name right) "id") "id")
			 (as-stm (get-:: (get-meta-name right) "attribute") "attribute")
			 (as-stm (get-:: (get-meta-name right) "value") "value")))))
      (code-append pig
		   (assignment-stm
		     (get-exp-name left)
		     (foreach-stm 
		       (par (join-stm (get-exp-name right) "id" (get-ids-name left) "id" "replicated"))
		       (comma-separated-list
			 (as-stm (get-:: (get-exp-name right) "id") "id")
			 (as-stm (get-:: (get-exp-name right) "region") "region")
			 (as-stm (get-:: (get-exp-name right) "value") "value")))))
      (list preamble pig))))

;scm: schemata list
;left: variable
;condition: string
;right: variable
(define (emit-select-code scm left condition right)
  (letrec ((right-schema-list (scm 'get (variable-identifier right) (variable-user right)))
	   (right-schema-struct (cond
				  ((equal? (length right-schema-list) 0) (not-existing-schema-error right))
				  ((> (length right-schema-list) 1) (multiple-schema-definition-error scm right))
				  (else  (first right-schema-list))))
	   (right-schema (second right-schema-struct))
	   (right-header (first right-schema-struct))
	   (right-indexed? (schemata-manager:is-indexed? right-header))
	   (left-name (variable-identifier left))
	   (left-line (variable-line-number left))
	   (right-type (schemata-manager:get-type right-schema-struct))
	   (right-user (schemata-manager:get-user right-header)))
    (begin
      (if 
	(null? (scm 'get left-name 'any)) 
	(scm 'add left-name right-schema)
	(variable-redefinition-error left))
      (if 
	right-indexed?
	(emit-lucene-select-code left right condition right-type right-user)
	(emit-select-code-simple left right condition right-user)))))

(define (emit-repository-select-all-code left right right-type right-user)
  (displayln right-user)
  (letrec
    ((pig "\n\n--PIG repositor SELECT ALL\n")
     (preamble "")
     (file-name (string-append repository-root "/" right-user "/datasets/" (variable-identifier right) ".xml"))
     (file-list (string-append "'" (string-join (map (λ(x) (string-append region-prefix x)) (xml-utils:xml->file-list (xml-utils:file->xexpr file-name))) ",") "'") )
     (schema "(id:long, region:(chr:chararray, left:int, right:int, strand:chararray), value:())")
     (loading-function "PigInputOutput.LoadRegions"))
    (begin
      (code-append pig
		   (assignment-stm
		     (get-exp-name left)
		     (string-append
		       "load "
		       file-list
		       " using "
		       loading-function
		       "(" "'" right-type "'"",""'" file-name "'"")"
		       " AS " schema)))
      (code-append pig
		   (assignment-stm
		     (get-meta-name left)
		     (foreach-stm
		       (par
			 (join-stm
			   (string-append right-user "_" (variable-identifier right) "_meta") "id"
			   (par (distinct-stm (foreach-stm (get-exp-name left) "id"))) "id"
			   "replicated"))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "attribute")
			 (as-stm "$2" "value")))))


      (list preamble pig))))

(define (emit-simple-select-all-code left right right-user)
  (letrec
    ((pig "")
     (preamble ""))
    (begin
      (code-append pig 
		   (assignment-stm
		     (get-exp-name left)
		     (get-exp-name right)))
      (code-append pig
		   (assignment-stm
		     (get-meta-name left)
		     (get-meta-name right)))
      (list preamble pig))))

;implement the translation for statements such as SELECT(*)
(define (emit-select-all-code scm left right)
  (letrec 
    ((right-schema-list (scm 'get (variable-identifier right) (variable-user right)))
     (right-schema-struct (cond
			    ((equal? (length right-schema-list) 0) (not-existing-schema-error right))
			    ((> (length right-schema-list) 1) (multiple-schema-definition-error scm right))
			    (else  (first right-schema-list))))
     (right-schema (second right-schema-struct))
     (right-header (first right-schema-struct))
     (right-indexed? (schemata-manager:is-indexed? right-header))
     (left-name (variable-identifier left))
     (left-line (variable-line-number left))
     (right-type (schemata-manager:get-type right-schema-struct))
     (right-user (schemata-manager:get-user right-header)))
    (begin
      (if
	(null? (scm 'get left-name 'any))
	(scm 'add left-name right-schema)
	(variable-redefinition-error left))
      (if
	right-indexed?
	(emit-repository-select-all-code left right right-type right-user)
	(emit-simple-select-all-code left right right-user)))))

(define (translate-attribute-modifier scm var mod)
  (cond
   ((and (list? mod) (equal? (length mod) 1)) (translate-attribute-modifier scm var (first mod)))
   ((and (list? mod) (equal? (length mod) 2)) (translate-attribute-modifier scm var (second mod)))
   ((and (list? mod) (equal? (length mod) 3)) (string-append "("
                                               (translate-attribute-modifier scm var (second mod)) " "
                                               (translate-attribute-modifier scm var (first mod)) " "
                                               (translate-attribute-modifier scm var (third mod)) ")"))
   ((string? mod) mod)
   ((equal? mod 'LEFT) "$1.$1")
   ((equal? mod 'RIGHT) "$1.$2")
   ((field-instance? mod) 
    (letrec
        ((arg-name (field-instance-identifier mod))
         (arg-path (field-instance-path mod))
         (arg-line (field-instance-line-number mod))
         (pos-list (scm 'position (variable-identifier var) (variable-user var) arg-name arg-path))
         (pos (cond
               ((equal? (length pos-list) 1) (first pos-list))
               ((equal? (length pos-list) 0) (not-existing-field-error arg-name arg-line (variable-identifier var)))
               ((> (length pos-list) 1) (multiple-field-definition arg-name arg-line (variable-identifier var)))))
         (type (schemata-manager:schema-field-type (list-ref (cadar (scm 'get (variable-identifier var) (variable-user var))) pos))))
      (begin
        (when 
            (not (or (equal? type 'INTEGER) (equal? type 'FLOAT) (equal? type 'INT) (equal? type 'DOUBLE)))
          (error (string-append 
                  "Field " (field-instance-identifier mod) 
                  " at line " (number->string (variable-line-number mod)) 
                  " is not allowed to appear in a project attribute extension. Only numeric type fields are permitted.")))
      (string-append "(double)$2.$" (number->string pos)))))))
  

;scm: schemata list
;left: variable
;condition: composed structure
;region-modifier: (list (list start op offset) (list start op offset))
;attribute-modifier: list that expresses the condition tree
;right: variable
(define (emit-project-code scm left expr region-modifier attribute-modifier right)
  (letrec 
    ((right-schema-list (scm 'get (variable-identifier right) (variable-user right)))
     (right-schema-struct (cond
			    ((equal? (length right-schema-list) 0) (not-existing-schema-error right))
			    ((> (length right-schema-list) 1) (multiple-schema-definition-error scm right))
			    (else  (first right-schema-list))))
     (right-schema (second right-schema-struct))
     (right-identifier (variable-identifier right))
     (left-name (variable-identifier left))
     (left-line (variable-identifier left))
     (condition 'null)
     (preamble "")
     (pig (string-append "\n-- PROJECT " (identifier-of left) " " (identifier-of right) "\n")))
    (begin
      (if
	(null? (scm 'get left-name 'any))
	(scm 'add left-name 
             (append 
              right-schema
              (map (lambda (x) 
                     (schemata-manager:schema-field 
                      (if (string? (second x)) 'STRING 'FLOAT) 
                      (field-instance-identifier (first x)) 
                      (list))) 
                   attribute-modifier)))
	(variable-redefinition-error left))
      (set! condition (translate-expression scm expr right))
      (cond
       ((null? region-modifier) (code-append pig
                                             (assignment-stm
                                              (string-append (get-exp-name left) "_partial2")
                                              (filter-stm (get-exp-name right) condition))))
       ((and (list? region-modifier) (equal? 'FROM-FIELD (first region-modifier)))
        (letrec
            ((new-left (if 
                        (null? (second region-modifier))
                        "$1.$1"
                        (letrec
                            ((pos-list (scm 'position (variable-identifier right) (variable-user right) 
                                            (field-instance-identifier (second (second region-modifier)))
                                            (field-instance-path (second (second region-modifier)))))
			      (pos (cond
				     ((equal? (length pos-list) 1) (first pos-list))
				     ((equal? (length pos-list) 0) (not-existing-field-error 
                                                                    (field-instance-identifier (second (second region-modifier)))
                                                                    (field-instance-line-number (second (second region-modifier)))
                                                                    (variable-identifier right)))
				     ((> (length pos-list) 1) (multiple-field-definition 
                                                               (field-instance-identifier (second (second region-modifier)))
                                                               (field-instance-line-number (second (second region-modifier)))
                                                               (variable-identifier right))))))
                          (string-append "$2.$" (number->string pos)))))
             (new-right (if 
                        (null? (third region-modifier))
                        "$1.$2"
                        (letrec
                            ((pos-list (scm 'position (variable-identifier right) (variable-user right) 
                                            (field-instance-identifier (second (third region-modifier)))
                                            (field-instance-path (second (third region-modifier)))))
			      (pos (cond
				     ((equal? (length pos-list) 1) (first pos-list))
				     ((equal? (length pos-list) 0) (not-existing-field-error 
                                                                    (field-instance-identifier (second (third region-modifier)))
                                                                    (field-instance-line-number (second (third region-modifier)))
                                                                    (variable-identifier right)))
				     ((> (length pos-list) 1) (multiple-field-definition 
                                                               (field-instance-identifier (second (third region-modifier)))
                                                               (field-instance-line-number (second (third region-modifier)))
                                                               (variable-identifier right))))))
                          (string-append "$2.$" (number->string pos))))))
          (code-append pig 
                       (assignment-stm
                        (string-append (get-exp-name left) "_partial2")
                        (filter-stm
                         (par
                          (foreach-stm 
                           (get-exp-name right) 
                           (comma-separated-list
                            "$0"
                            (string-append "TOTUPLE($1.$0,(int)" new-left ",(int)" new-right ",$1.$3)")
                            "$2")))
                         (string-append condition " AND $1.$1 < $1.$2"))))))
       (else (begin
               (code-append pig
                            (assignment-stm
                             (string-append (get-exp-name left) "_partial1")
                             (filter-stm (get-exp-name right) condition)))
               (code-append pig (define-region-modifier left region-modifier))
               (code-append pig
                            (assignment-stm
                             (string-append (get-exp-name left) "_partial2")
                             (foreach-stm
                              (par
                               (filter-field-not-null
                                (par
				 (foreach-stm
                                  (string-append (get-exp-name left) "_partial1")
                                  (comma-separated-list
                                   "$0"
                                   (apply-function
                                    (string-append (identifier-of left) "_modifier")
                                    (comma-separated-list  "$1.$0" "$1.$1" "$1.$2" "$1.$3")) 
                                   "$2")))
                                "$1"))
                              (comma-separated-list
                               (as-stm "$0" "id")
                               (as-stm "$1" "region")
                               (as-stm "$2" "value"))))))))
      (if
       (null? attribute-modifier)
       (code-append pig 
                    (assignment-stm
                     (get-exp-name left)
                     (string-append (get-exp-name left) "_partial2")))
       (code-append pig 
                    (assignment-stm
                     (get-exp-name left)
                     (foreach-stm
                     (par
                      (foreach-stm
                      (string-append (get-exp-name left) "_partial2")
                      (comma-separated-list
                       "$0" "$1"
                       (apply-function
                        "TOTUPLE"
                       (string-join
                        (append
                         (map
                          (lambda (x) (string-append "$2.$" (number->string x))) 
                          (range (length right-schema)))
                         (map
                          (lambda (x) (translate-attribute-modifier scm right (second x)))
                          attribute-modifier)) ",")))))
                     (comma-separated-list
                      (as-stm "$0" "id")
                      (as-stm "$1" "region")
                      (as-stm "$2" "value"))))))
       
      (code-append pig
		   (assignment-stm
                    (get-ids-name left)
                    (distinct-stm 
                     (foreach-stm (get-exp-name left) "id"))))
      (code-append pig
		   (assignment-stm
		     (get-meta-name left)
		     (foreach-stm
		       (par (join-stm (get-meta-name right) "id" (get-ids-name left) "id" "replicated"))
		       (comma-separated-list
			 (as-stm (get-:: (get-meta-name right) "id") "id")
			 (as-stm (get-:: (get-meta-name right) "attribute") "attribute")
			 (as-stm (get-:: (get-meta-name right) "value") "value")))))
      (list preamble pig)
      )))

;scm: schemata
;left: variable
;function: list of strings
;referiment: variable
;dataset: variable
;type: string ("stranded" | "not-stranded")
(define (emit-map-code scm left type meta-filter function reference dataset)
  (letrec
    ((reference-schema-list (scm 'get (variable-identifier reference) (variable-user reference)))
     (reference-schema-struct (cond
                               ((equal? (length reference-schema-list) 0) (not-existing-schema-error reference))
				((> (length reference-schema-list) 1) (multiple-schema-definition-error scm reference))
				(else  (first reference-schema-list))))
     (reference-schema (second reference-schema-struct))
     (dataset-schema-list (scm 'get (variable-identifier dataset) (variable-user dataset)))
     (dataset-schema-struct (cond
			      ((equal? (length dataset-schema-list) 0) (not-existing-schema-error dataset))
			      ((> (length dataset-schema-list) 1) (multiple-schema-definition-error scm dataset))
			      (else  (first dataset-schema-list))))
     (dataset-schema (second dataset-schema-struct))
     (preamble "")
     (left-name (variable-identifier left))
     (reference-name (variable-identifier reference))
     (dataset-name (variable-identifier dataset))
     (pig (string-append "\n-- MAP " (identifier-of left) " " (identifier-of reference) " " (identifier-of dataset) "\n")))
    (begin
        (if 
	(null? (scm 'get left-name 'any))
	(let
	  ((new-schema reference-schema))
	  (begin
	    (let add-schema ((l function))
	      (when (not (null? l))
		(let ((curf (car l)))
		  (cond
                    ((equal? (length curf) 2) (set! new-schema (schemata-manager:append-field-to-schema new-schema (first curf) 'INTEGER)))
		    ((equal? (length curf) 3)
		     (letrec ((func-name (second curf))
			      (arg (third curf))
			      (arg-name (field-instance-identifier arg))
			      (arg-path (field-instance-path arg))
			      (arg-line (field-instance-line-number arg))
                              (new-field-name (first curf))
                              (new-field-type (cond
                                               ((or (equal? func-name "MIN")
                                                    (equal? func-name "MAX")
                                                    (equal? func-name "SUM")
                                                    (equal? func-name "AVG")) 'FLOAT)
                                               ((equal? func-name "BAG") 'STRING)))
			      (pos-list (scm 'position (variable-identifier dataset) (variable-user dataset) arg-name arg-path))
			      (pos (cond
				     ((equal? (length pos-list) 1) (first pos-list))
				     ((equal? (length pos-list) 0) (not-existing-field-error arg-name arg-line (variable-identifier dataset)))
				     ((> (length pos-list) 1) (multiple-field-definition arg-name arg-line (variable-identifier dataset))))))
		       (set! new-schema (schemata-manager:append-field-to-schema new-schema (string-append new-field-name "_" arg-name) new-field-type))))
		    (else (error "unknown aggregate function in MAP")))
		  (add-schema (cdr l)))))
	    (scm 'add left-name new-schema)))
	(variable-redefinition-error left))
      (code-append pig
		   (let ;if this point is reached, the field exists and it is unique, no need to check
		     ((arg-list (list)))
		     (let loop ((l function))
		       (when (not (null? l))
			 (let
			   ((cur (car l)))
			   (begin
			     (if (equal? (length cur) 2)
			       (set! arg-list (append arg-list (list (second cur))))
			       (let ((pos  (number->string (first (scm 'position
								       (variable-identifier dataset)
								       (variable-user dataset)
								       (field-instance-identifier (third cur))
								       (field-instance-path (third cur)))))))
				 (set! arg-list (append arg-list (list (second cur) pos)))))
			     (loop (cdr l))))))
		     (define-map-function left arg-list type)))
      (code-append pig
		   (assignment-stm
		     (get-exp-group-name dataset)
		     (group-stm
		       (get-exp-name dataset)
		       (par (comma-separated-list "$0" (accessor "$1" "$0"))))))
      (code-append pig
		   (assignment-stm
		     (get-exp-group-name reference)
		     (group-stm
		       (get-exp-name reference)
		       (par (comma-separated-list "$0" (accessor "$1" "$0"))))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name left) "_all_ids")
		     (cross-stm
		       (par
			 (distinct-stm
			   (foreach-stm
			     (get-exp-name reference)
			     "$0")))
		       (par
			 (distinct-stm
			   (foreach-stm
			     (get-exp-name dataset)
			     "$0"))))))
      (cond
	((symbol? meta-filter) (code-append pig
					    (assignment-stm
					      (string-append (get-exp-name left) "_allowed_ids")
					      (string-append (get-exp-name left) "_all_ids"))))
	((string? meta-filter) (begin
				 (code-append pig 
					      (define-join-meta-filter left meta-filter))
				 (code-append pig
					      (assignment-stm
						(get-meta-group-name reference)
						(group-stm
						  (get-meta-name reference) "$0")))
				 (code-append pig
					      (assignment-stm
						(get-meta-group-name dataset)
						(group-stm
						  (get-meta-name dataset) "$0")))
				 (code-append pig
					      (assignment-stm
						(string-append (get-exp-name left) "_allowed_ids1")
						(join-stm
						  (string-append (get-exp-name left) "_all_ids") "$0"
						  (get-meta-group-name reference) "$0")))
				 (code-append pig
					      (assignment-stm
						(string-append (get-exp-name left) "_allowed_ids")
						(foreach-stm
						  (par
						    (filter-stm
						      (par
							(foreach-stm
							  (par
							    (join-stm
							      (string-append (get-exp-name left) "_allowed_ids1") "$1"
							      (get-meta-group-name dataset) "$0"))
							  (comma-separated-list
							    "$0"
							    "$1"
							    (apply-function
							      (get-meta-filter-function-name left) (par (comma-separated-list "$3" "$5"))))))
						      "$2==true"))
						  (comma-separated-list 
						    "$0"
						    "$1")))))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name left) "_chr_groups")
		     (nested-foreach-stm
		       (par
			 (group-stm
			   (par
			     (distinct-stm
			       (foreach-stm
				 (get-exp-name reference)
				 (comma-separated-list "$0" "$1.$0"))))
			   "$0"))
		       (list (assignment-stm "chrs" (foreach-stm "$1" "$1")))
		       (comma-separated-list "$0" "chrs"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name left) "_partial1")
		     (foreach-stm
		       (par
			 (foreach-stm
			   (par
			     (join-stm
			       (string-append (get-exp-name left) "_allowed_ids") "$0"
			       (string-append (get-exp-name left) "_chr_groups") "$0"))
			   (comma-separated-list "$0" "$1" "flatten($3)")))
		       (comma-separated-list 
			 (par (comma-separated-list "$0" "$2"))
			 (par (comma-separated-list "$1" "$2"))))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name left) "_partial2")
		     (foreach-stm
		       (par
			 (join-stm
			   (string-append (get-exp-name left) "_partial1") "$0"
			   (get-exp-group-name reference) "$0"))
		       (comma-separated-list "$0" "$1" "$3"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name left) "_partial3")
		     (foreach-stm
		       (par
			 (join-stm
			   (string-append (get-exp-name left) "_partial2") "$1 left"
			   (get-exp-group-name dataset) "$0"))
		       (comma-separated-list "$2" "$4" "$1.$0"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name left) "_map")
		     (foreach-stm
		       (string-append (get-exp-name left) "_partial3")
		       (comma-separated-list "$2" (apply-function
						    (get-mapper-function-name left)
						    "TOTUPLE($0,$1)")))))
      (code-append pig
		   (assignment-stm
		     (get-exp-name left)
		     (foreach-stm
		       (par
			 (foreach-stm
			   (string-append(get-exp-name left) "_map")
			   (comma-separated-list "$0" "flatten($1)")))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$2" "region")
			 (as-stm "$3" "value")))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-meta-name left) "_ref")
		     (foreach-stm
		       (par
			 (join-stm
			   (string-append (get-exp-name left) "_allowed_ids") "$0"
			   (get-meta-name reference) "$0"))
		       (comma-separated-list
			 "$1"
			 (apply-function
			   "CONCAT"
			   (comma-separated-list 
			     (string-append "'" (variable-identifier reference) ".'")
			     "$3"))
			 "$4"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-meta-name left) "_ds")
		     (foreach-stm
		       (get-meta-name dataset)
		       (comma-separated-list
			 "$0"
			 (apply-function
			   "CONCAT"
			   (comma-separated-list
			     (string-append "'" (variable-identifier dataset) ".'")
			     "$1"))
			 "$2"))))
      (code-append pig
		   (assignment-stm
		     (get-meta-name left)
		     (foreach-stm
		       (par
			 (union-stm
			   (string-append (get-meta-name left) "_ref")
			   (string-append (get-meta-name left) "_ds")))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "attribute")
			 (as-stm "$2" "value")))))
      (list preamble pig))))

;scm: schemata
;var: variable to materialize
(define (emit-materialize-code scm var)
  (letrec
    ((schema-list (scm 'get (variable-identifier var) (variable-user var)))
     (schema-struct (cond
		      ((equal? (length schema-list) 0) (not-existing-schema-error var))
		      ((> (length schema-list) 1) (multiple-schema-definition-error scm var))
		      (else  (first schema-list))))
     (preamble "")
     (pig ""))
    (begin
      (set! pig (string-append pig (store-stm var)))
      (list preamble pig))))


;scm: schemata closure
;join-args: a list of arguments
;join-type; string, either "stranded" or "not-stranded"
(define (emit-join-code scm new-var meta-cond geno-cond comp type left right)
  (letrec
    ((left-schema-list (scm 'get (variable-identifier left) (variable-user left)))
     (left-schema-struct (cond
			   ((equal? (length left-schema-list) 0) (not-existing-schema-error left))
			   ((> (length left-schema-list) 1) (multiple-schema-definition-error scm left))
			   (else  (first left-schema-list))))
     (left-schema (second left-schema-struct))
     (right-schema-list (scm 'get (variable-identifier right) (variable-user right)))
     (right-schema-struct (cond
			    ((equal? (length right-schema-list) 0) (not-existing-schema-error right))
			    ((> (length right-schema-list) 1) (multiple-schema-definition-error scm right))
			    (else  (first right-schema-list))))
     (right-schema (second right-schema-struct))
     (new-var-name (variable-identifier new-var))
     (left-name (variable-identifier left))
     (right-name (variable-identifier right))
     (preamble "")
     (pig (string-append "\n-- JOIN " (identifier-of new-var) " " (identifier-of left) " " (identifier-of right) "\n")))
    (begin
      (if
	(null? (scm 'get new-var-name 'any))
	(cond
	  ((equal? comp "r1_distinct") (scm 'add new-var-name 
					    (append
                                             (schemata-manager:merge-two-schemata 
					      left-schema left-name 
                                              (map (λ (x) 
                                                     (schemata-manager:schema-field 
                                                      'STRING 
                                                      (string-append (schemata-manager:schema-field-name x) "_aggregation") 
                                                      (schemata-manager:schema-field-full-path x)))
                                                   right-schema)
                                              right-name)
                                             (list (schemata-manager:schema-field 'STRING "distance_aggregation" (list))))))
	  ((equal? comp "r2_distinct") (scm 'add new-var-name 
                                            (append 
                                             (schemata-manager:merge-two-schemata 
                                              (map (λ (x)
                                                     (schemata-manager:schema-field
                                                      'STRING
                                                      (string-append (schemata-manager:schema-field-name x) "_aggregation")
                                                      (schemata-manager:schema-field-full-path x)))
                                                   left-schema)
                                              left-name 
					      right-schema 
                                              right-name)
                                             (list (schemata-manager:schema-field 'STRING "distance_aggregation" (list))))))
	  ((equal? comp "project_r1") (scm 'add new-var-name left-schema))
	  ((equal? comp "project_r2") (scm 'add new-var-name right-schema))
	  ((equal? comp "project_r1_distinct") (scm 'add new-var-name left-schema))
	  ((equal? comp "project_r2_distinct") (scm 'add new-var-name right-schema))
	  (else (scm 'add new-var-name (schemata-manager:append-field-to-schema 
					 (schemata-manager:merge-two-schemata left-schema left-name right-schema right-name)
					 "distance" 'INTEGER))))
	(variable-redefinition-error new-var)) ;else..
      (code-append pig
		   (assignment-stm
		     (get-exp-group-name left)
		     (group-stm (get-exp-name left) "$0")))
      (code-append pig
		   (assignment-stm
		     (get-exp-group-name right)
		     (group-stm (get-exp-name right) "$0")))
      (code-append pig
		   (assignment-stm
		     (get-meta-group-name left)
		     (group-stm (get-meta-name left) "$0")))
      (code-append pig
		   (assignment-stm
		     (get-meta-group-name right)
		     (group-stm (get-meta-name right) "$0")))
      (code-append pig
		   (assignment-stm
		     (get-exp-cross-name left right)
		     (foreach-stm 
		       (par
			 (cross-stm (get-exp-group-name left) (get-exp-group-name right)))
		       (comma-separated-list
			 (par
			   (comma-separated-list "$0" "$2"))
			 (par
			   (comma-separated-list "$1" "$3"))))))
      (code-append pig
		   (assignment-stm
		     (get-meta-cross-name left right)
		     (foreach-stm 
		       (par	
			 (cross-stm (get-meta-group-name left) (get-meta-group-name right)))
		       (comma-separated-list
			 (par
			   (comma-separated-list "$0" "$2"))
			 (par
			   (comma-separated-list "$1" "$3"))))))
      (code-append pig
		   ; It passes '(genomitric condition, composition, stranded)
		   (let
		     ((arg-list (append geno-cond (list comp) (list type))))
		     (define-join-function new-var arg-list)))
      (when (string? meta-cond)
	(begin
	  (code-append pig
		       (define-join-meta-filter new-var meta-cond))
	  (code-append pig
		       (assignment-stm
			 (get-ids-name new-var)
			 (foreach-stm
			   (par
			     (filter-stm
			       (par
				 (foreach-stm
				   (get-meta-cross-name left right)
				   (comma-separated-list "$0" (apply-function (get-meta-filter-function-name new-var) "$1"))))
			       "$1==true"))
			   "$0")))
	  (code-append pig
		       (assignment-stm
			 (get-flattened-name (get-meta-cross-name left right))
			 (foreach-stm
			   (par
			     (join-stm
			       (get-meta-cross-name left right)
			       "$0"
			       (get-ids-name new-var)
			       "$0"))
			   (comma-separated-list "$0" "flatten($1)" ))))
	  (code-append pig
		       (assignment-stm
			 (get-meta-name new-var)
			 (foreach-stm
			   (par
			     (union-stm
			       (par
				 (foreach-stm
				   (par
				     (foreach-stm
				       (get-flattened-name (get-meta-cross-name left right))
				       (comma-separated-list 
					 (apply-function
					   "GenometricPig.NewId" 	
					   (comma-separated-list
					     (accessor "$0" "$0")
					     (accessor "$0" "$1")))
					 "flatten($1)")))
				   (comma-separated-list 
				     "$0" 
				     (string-append "CONCAT('" (variable-identifier left) ".',$2)")
				     "$3")))
			       (par
				 (foreach-stm
				   (par
				     (foreach-stm
				       (get-flattened-name (get-meta-cross-name left right))
				       (comma-separated-list
					 (apply-function
					   "GenometricPig.NewId"
					   (comma-separated-list
					     (accessor "$0" "$0")
					     (accessor "$0" "$1")))
					 "flatten($2)")))
				   (comma-separated-list
				     "$0"
				     (string-append "CONCAT('" (variable-identifier right) ".',$2)")
				     "$3")))))
			   (comma-separated-list
			     (as-stm "$0" "id")
			     (as-stm "$1" "attribute")
			     (as-stm "$2" "value")))))
	  (code-append pig
		       (assignment-stm
			 (get-exp-name new-var)
			 (foreach-stm
			   (par
			     (foreach-stm
			       (par
				 (foreach-stm
				   (par
				     (join-stm
				       (get-exp-cross-name left right)
				       "$0"
				       (get-ids-name new-var)
				       "$0"))
				   (apply-function (get-joiner-function-name new-var) "$1")))
			       "flatten($0)"))
			   (comma-separated-list
			     (as-stm "$0" "id")
			     (as-stm "$1" "region")
			     (as-stm "$2" "value")))))
	  ))
      (unless (string? meta-cond)
	(begin
	  (code-append pig
		       (assignment-stm
			 (get-flattened-name (get-meta-cross-name left right))
			 (foreach-stm
			   (get-meta-cross-name left right)
			   (comma-separated-list "$0" "flatten($1)" ))))
	  (code-append pig
		       (assignment-stm
			 (get-meta-name new-var)
			 (foreach-stm
			   (par
			     (union-stm
			       (par
				 (foreach-stm
				   (get-flattened-name (get-meta-cross-name left right))
				   (comma-separated-list 
				     (apply-function
				       "GenometricPig.NewId" 	
				       (comma-separated-list
					 (accessor "$0" "$0")
					 (accessor "$0" "$1")))
				     "flatten($1)")))
			       (par
				 (foreach-stm
				   (get-flattened-name (get-meta-cross-name left right))
				   (comma-separated-list
				     (apply-function
				       "GenometricPig.NewId"
				       (comma-separated-list
					 (accessor "$0" "$0")
					 (accessor "$0" "$1")))
				     "flatten($2)")))))
			   (comma-separated-list
			     (as-stm "$0" "id")
			     (as-stm "$2" "attribute")
			     (as-stm "$3" "value")))))
	  (code-append pig
		       (assignment-stm
			 (get-meta-name new-var)
			 (foreach-stm
			   (par
			     (union-stm
			       (par
				 (foreach-stm
				   (par
				     (foreach-stm
				       (get-flattened-name (get-meta-cross-name left right))
				       (comma-separated-list 
					 (apply-function
					   "GenometricPig.NewId" 	
					   (comma-separated-list
					     (accessor "$0" "$0")
					     (accessor "$0" "$1")))
					 "flatten($1)")))
				   (comma-separated-list
				     "$0"
				     (string-append "CONCAT('" (variable-identifier left) ".',$2)")
				     "$3")))
			       (par
				 (foreach-stm
				   (par
				     (foreach-stm
				       (get-flattened-name (get-meta-cross-name left right))
				       (comma-separated-list
					 (apply-function
					   "GenometricPig.NewId"
					   (comma-separated-list
					     (accessor "$0" "$0")
					     (accessor "$0" "$1")))
					 "flatten($2)")))
				   (comma-separated-list
				     "$0"
				     (string-append "CONCAT('" (variable-identifier right) ".',$2)")
				     "$3")))))
			   (comma-separated-list
			     (as-stm "$0" "id")
			     (as-stm "$1" "attribute")
			     (as-stm "$2" "value")))))
	  (code-append pig
		       (assignment-stm
			 (get-exp-name new-var)
			 (foreach-stm
			   (par
			     (foreach-stm
			       (par
				 (foreach-stm
				   (get-exp-cross-name left right)
				   (apply-function (get-joiner-function-name new-var) "$1")))
			       "flatten($0)"))
			   (comma-separated-list
			     (as-stm "$0" "id")
			     (as-stm "$1" "region")
			     (as-stm "$2" "value")))))
	  ))
      (list preamble pig))))


;new-var is a variable
;operand-list is the list of variable to be united
(define (emit-union-code scm new-var operand-list )
  (letrec
    ((new-var-name (variable-identifier new-var))
     (ref-dataset (first operand-list))
     (ref-schema-list (scm 'get (variable-identifier ref-dataset) (variable-user ref-dataset)))
     (ref-schema-struct (cond
			  ((equal? (length ref-schema-list) 0) (not-existing-schema-error ref-dataset))
			  ((> (length ref-schema-list) 1) (multiple-schema-definition-error scm ref-dataset))
			  (else  (first ref-schema-list))))
     (ref-schema (second ref-schema-struct))
     (preamble "")
     (pig ""))
    (begin
      (if
	(null? (scm 'get new-var-name 'any))
	(scm 'add new-var-name ref-schema)
	(variable-redefinition-error new-var))
      (let loop ((l (cdr operand-list)) (counter 1))
	(begin
	  (letrec
	    ((var (car l))
	     (var-name (variable-identifier var))
	     (var-owner (variable-user var))
	     (var-schema-list (scm 'get (variable-identifier var) (variable-user var)))
	     (var-schema-struct (cond
				  ((equal? (length var-schema-list) 0) (not-existing-schema-error var))
				  ((> (length var-schema-list) 1) (multiple-schema-definition-error var))
				  (else (first var-schema-list))))
	     (var-schema (second var-schema-struct)))
	    (let
	      ((args
		 (string-join
		   (map
		     (λ (x) 
			(letrec
			  ((ref-field-name (schemata-manager:schema-field-name x))
			   (ref-field-path (schemata-manager:schema-field-full-path x))
			   (pos-list (scm 'position var-name var-owner ref-field-name ref-field-path)))
			  (if (equal? (length pos-list) 1)
			    (string-append "$2.$" (number->string (first pos-list)))
			    "null")))
		     ref-schema)
		   ",")))
	      (code-append pig
			   (assignment-stm
			     (string-append (get-exp-name var) "_partial")
			     (foreach-stm
			       (get-exp-name var)
			       (comma-separated-list
				 (apply-function
				   "GenometricPig.NewId"
				   (comma-separated-list
				     (string-append (number->string counter) "L")
				     "$0"))
				 "$1"
				 (string-append "TOTUPLE("args")"))))))

	    (code-append pig   ;metadata genaration
			 (assignment-stm
			   (string-append (get-meta-name var) "_partial")
			   (foreach-stm
			     (get-meta-name var)
			     (comma-separated-list
			       (apply-function
				 "GenometricPig.NewId"
				 (comma-separated-list
				   (string-append (number->string counter) "L")
				   "$0"))
			       "$1"
			       "$2"))))
	    )
	  (when (not (null? (cdr l))) (loop (cdr l) (+ counter 1)))))
      (code-append pig
		   (assignment-stm 
		     (get-exp-name new-var)
		     (foreach-stm
		       (par
			 (string-append
			   "union "
			   (comma-separated-list
			     (par
			       (foreach-stm
				 (get-exp-name (first operand-list))
				 (comma-separated-list
				   (apply-function
				     "GenometricPig.NewId"
				     (comma-separated-list "0L" "$0"))
				   "$1" "$2")))
			     (string-join (map (λ (x) (string-append (get-exp-name x) "_partial")) (cdr operand-list)) ","))))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "region")
			 (as-stm "$2" "value")))))

      (code-append pig
		   (assignment-stm
		     (get-meta-name new-var)
		     (foreach-stm
		       (par
			 (string-append
			   "union "
			   (comma-separated-list
			     (par
			       (foreach-stm
				 (get-meta-name (first operand-list))
				 (comma-separated-list
				   (apply-function
				     "GenometricPig.NewId"
				     (comma-separated-list "0L" "$0"))
				   "$1"
				   "$2")))
			     (string-join (map (λ (x) (string-append  (get-meta-name x) "_partial")) (cdr operand-list)) ","))))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "attribute")
			 (as-stm "$2" "value")))))
      (list preamble pig))))

;left: variable, new defined variable
;type: symbol, 'cover, 'cover-flat, 'summit, summit-flat
;minAcc: string
;maxAcc: string
;jaccard-min: string
;jaccard-max:string
;groups: (eventually empty) string list of group by names
;merge: boolean
;input-datasets; (list input [mandatory-region])
(define (emit-region-intersection-code scm left-var type minAcc maxAcc jaccard-min jaccard-max groups merge input-datasets agg-funtion-list)
  (letrec  
    ((left-var-name (variable-identifier left-var))
     (input-dataset (first input-datasets))
     (mandatory (when (equal? (length input-datasets) 2) (second input-datasets)))
     (input-schema-list (scm 'get (variable-identifier input-dataset) (variable-user input-dataset)))
     (input-schema-struct (cond
			    ((equal? (length input-schema-list) 0) (not-existing-schema-error input-dataset))
			    ((> (length input-schema-list) 1) (multiple-schema-definition-error scm input-dataset))
			    (else  (first input-schema-list))))
     (input-schema (second input-schema-struct))
     (mandatory-dataset (if (equal? (length input-datasets) 2) (second input-datasets) 'none))
     (mandatory-schema-list (unless (symbol? mandatory-dataset) (scm 'get (variable-identifier mandatory-dataset) (variable-user mandatory-dataset))))
     (mandatory-schema-struct (unless (symbol? mandatory-dataset)
				(cond
				  ((equal? (length mandatory-schema-list) 0) (not-existing-schema-error mandatory-dataset))
				  ((> (length mandatory-schema-list) 1) (multiple-schema-definition-error scm mandatory-dataset))
				  (else  (first mandatory-schema-list)))))
     (has-groups? (not (null? groups)))
     (has-mandatory? (equal? (length input-datasets) 2))
     (agg-func-parameters (string-join 
                           (map 
                           (lambda (f)
                             (if
                              (or (equal? (first f) "COUNT") (equal? (first f) "EXISTS"))
                              (string-append (first f) " -1")
                              (letrec 
                                  ((var-name (variable-identifier input-dataset))
                                   (var-user (variable-user input-dataset))
                                   (field-name (field-instance-identifier (third f)))
                                   (field-path (list)) ;this should be substituted with the correct value
                                   (field-number (field-instance-line-number (third f)))
                                   (field-position 
                                    (let ((pos-list (scm 'position var-name var-user field-name field-path)))
                                      (cond
                                       ((equal? (length pos-list) 0) (not-existing-field-error field-name field-number var-name))
                                       ((> (length pos-list) 1) (multiple-field-definition field-name field-number var-name))
                                       (else (car pos-list))))))
                                (string-append
                                 (first f) " "
                                 (number->string field-position)))))
                           agg-funtion-list) " "))
     (preamble "")
     (pig ""))
    (begin
      (if
	(null? (scm 'get left-var-name 'any))
	(scm 'add left-var-name (append 
                                 (list (schemata-manager:schema-field 'FLOAT "JaccardIndex" (list)))
                                 (map
                                  (lambda (f) (schemata-manager:schema-field
                                          (if (equal? (first f) 'BAG) 'STRING 'FLOAT)
                                          (field-instance-identifier (second f)) (list)))
                                  agg-funtion-list)))
	(variable-redefinition-error left-var))
      (code-append pig
		   (define-intersection-operation left-var type minAcc maxAcc jaccard-min jaccard-max merge agg-func-parameters))
      (code-append pig
		   (define-intersection-metadata-grouper left-var groups))
      (code-append pig
		   (assignment-stm
		     (get-ids-name input-dataset)
		     (foreach-stm
		       (par
			 (group-stm
			   (get-meta-name input-dataset)
			   "$0"))
		       (comma-separated-list
			 "$0"
			 (apply-function
			   (get-cover-metadata-grouper-name left-var)
			   "$1")))))
      (code-append pig
		   (assignment-stm
		     (string-append (identifier-of left-var) "_counting")
		     (foreach-stm
		       (par 
			 (group-stm
			   (get-ids-name input-dataset)
			   "$1"))
		       (comma-separated-list
			 "$0"
			 (apply-function
			   "COUNT"
			   "$1")))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name input-dataset) "_partial_1")
		     (foreach-stm
		       (par
			 (join-stm
			   (get-exp-name input-dataset) "$0"
			   (get-ids-name input-dataset) "$0"))
		       (comma-separated-list
			 "$4" "$1" "$2"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name input-dataset) "_partial")
		     (foreach-stm
		       (par
			 (join-stm
			   (string-append (get-exp-name input-dataset) "_partial_1") "$0"
			   (string-append (identifier-of left-var) "_counting") "$0"))
		       (comma-separated-list
			 "$0" "$1" "$2" "$4"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name input-dataset) "_partial_group")
		     (nested-foreach-stm
		       (par
			 (group-stm
			   (string-append (get-exp-name input-dataset) "_partial") 
			   (par (comma-separated-list "$0" "$1.$0" "$3"))))
		       (list (assignment-stm
			       "clean"
			       (foreach-stm "$1" (comma-separated-list "$0" "$1" "$2"))))
		       (comma-separated-list "$0" "clean"))))

      (cond
	(has-mandatory? (begin
			  ;this feature is not implemented. Understanding how the UDF works is required. Pherhaps it is not possible
			  ;passing all the regions with the same ids, but different ids for each sample are needed.
			  (error "Fatal: in this release, it is not possible to use Cover/Summit functions with the set of mandatory regions. Sorry.")
			  (code-append pig
				       (assignment-stm
					 (string-append (get-exp-group-name mandatory-dataset) "_mandatory")
					 (group-stm
					   (get-exp-name mandatory-dataset)
					   "$1.$0")))
			  (code-append pig
				       (assignment-stm
					 (string-append (get-exp-name left-var) "_func_output")
					 (foreach-stm
					   (par
					     (foreach-stm
					       (par
						 (join-stm
						   (string-append (get-exp-name input-dataset) "_partial_group") "$0.$1"
						   (string-append (get-exp-group-name mandatory-dataset) "_mandatory") "$0"))
					       (comma-separated-list "$0" "$1" "$3")))
					   (comma-separated-list
					     "$0"
					     (apply-function
					       (get-intersection-function left-var)
					       (comma-separated-list "$0.$0" "$1" "$2"))))))
			  (code-append pig
				       (assignment-stm
					 (get-exp-name left-var)
					 (foreach-stm
					   (par
					     (foreach-stm
					       (string-append (get-exp-name left-var) "_func_output")
					       "flatten($1.$0)"))
					   (comma-separated-list
					     (as-stm "$0" "id")
					     (as-stm "$1" "region")
					     (as-stm "$2" "value")))))
			  (code-append pig
				       (assignment-stm
					 (string-append (get-meta-name left-var) "_partial1")
					 (foreach-stm
					   (par
					     (join-stm
					       (get-meta-name input-dataset) "$0"
					       (get-ids-name input-dataset) "$0"
					       "replicated"))
					   (comma-separated-list "$4" "$1" "$2" ))))

			  (code-append pig
				       (assignment-stm
					 (string-append (get-meta-name left-var) "_partial2")
					 (foreach-stm
					   (par
					     (cross-stm
					       (par
						 (distinct-stm
						   (foreach-stm
						     (get-ids-name input-dataset)
						     "$1")))
					       (get-meta-name mandatory-dataset)))
					   (comma-separated-list 
					     "$0" 
					     (string-append "CONCAT('" (variable-identifier mandatory-dataset) ".',$2)") 
					     "$3"))))
			  (code-append pig
				       (assignment-stm
					 (get-meta-name left-var)
					 (foreach-stm
					   (par
					     (distinct-stm
					       (union-stm
						 (string-append (get-meta-name left-var) "_partial1")
						 (string-append (get-meta-name left-var) "_partial2")))) 
					   (comma-separated-list
					     (as-stm "$0" "id")
					     (as-stm "$1" "attribute")
					     (as-stm "$2" "value")))))))

	((not has-mandatory?) (begin
				(code-append pig
					     (assignment-stm
					       (string-append (get-exp-name left-var) "_func_output")
					       (foreach-stm
						 (par
						   (foreach-stm
						     (string-append (get-exp-name input-dataset) "_partial_group")
						     (comma-separated-list 
						       "$0"
						       (apply-function
							 (get-intersection-function left-var)
							 (comma-separated-list "$0.$0" "$0.$2" "$1")))))
						 (comma-separated-list
						   "$0" "$1.$0" "$1.$1" "$1.$2" "$1.$3" "$1.$4"))))
				(code-append pig
					     (assignment-stm
					       (get-exp-name left-var)
					       (foreach-stm
						 (par
						   (foreach-stm
						     (string-append (get-exp-name left-var) "_func_output")
						     "flatten($1)"))
						 (comma-separated-list
						   (as-stm "$0" "id")
						   (as-stm "$1" "region")
						   (as-stm "$2" "value")))))
                                (code-append pig
                                             (assignment-stm
                                              (string-append "new_" (get-meta-name left-var) "_part1")
                                              (foreach-stm
                                               (par
                                                (foreach-stm
                                                 (par
                                                  (group-stm
                                                   (par 
                                                    (foreach-stm
                                                     (string-append (get-exp-name left-var) "_func_output")
                                                     (comma-separated-list
                                                      "$0.$0" "$2" "$3" "$4" "$5")))
                                                   "$0"))
                                                 (comma-separated-list
                                                  "$0"
                                                  "TOBAG(TOTUPLE ($0, 'JaccardIndex', (chararray)((double)SUM($1.$1)/(double)SUM($1.$2))),TOTUPLE ($0, 'GlobalJaccardIndex', (chararray)((double)SUM($1.$3)/(double)SUM($1.$4))))")))
                                               (apply-function "flatten" "$1"))))
                                (code-append pig
					     (assignment-stm
					       (string-append "new_" (get-meta-name left-var) "_part2")
					       (foreach-stm
						 (par 
						   (distinct-stm
						     (foreach-stm
						       (par
							 (join-stm
							   (get-meta-name input-dataset) "$0"
							   (get-ids-name input-dataset) "$0"
							   "replicated"))
						       (comma-separated-list "$4" "$1" "$2" ))))
						 (comma-separated-list
						   (as-stm "$0" "id")
						   (as-stm "$1" "attribute")
						   (as-stm "$2" "value")))))
                                (code-append pig 
                                             (assignment-stm
                                              (get-meta-name left-var)
                                              (union-stm
                                               (string-append "new_" (get-meta-name left-var) "_part1")
                                               (string-append "new_" (get-meta-name left-var) "_part2")))))))

      (list preamble pig)
      )))

;new-var, to-be-filtered, filter-dataset: variable
;type: string, either "stranded" or "not-stranded"
;meta-filter: either a symbol (if no metadata filerign has been set) or a string (the filter to be passed to the function) 
(define (emit-difference-code scm new-var type meta-filter to-be-filtered filter-dataset)
  (letrec
    ((to-be-filtered-schema-list (scm 'get (variable-identifier to-be-filtered) (variable-user to-be-filtered)))
     (to-be-filtered-schema-struct (cond
				     ((equal? (length to-be-filtered-schema-list) 0) (not-existing-schema-error to-be-filtered))
				     ((> (length to-be-filtered-schema-list) 1) (multiple-schema-definition-error scm to-be-filtered))
				     (else  (first to-be-filtered-schema-list))))
     (to-be-filtered-schema (second to-be-filtered-schema-struct))
     (new-var-name (variable-identifier new-var))
     (to-be-filtered-name (variable-identifier to-be-filtered))
     (filter-dataset-name (variable-identifier filter-dataset))
     (pig "\n\n--DIFFERENCE\n")
     (preamble ""))
    (begin
      (if
	(null? (scm 'get new-var-name 'any))
	(scm 'add new-var-name to-be-filtered-schema)
	(variable-redefinition-error new-var))
      (code-append pig
		   (define-difference-function new-var type))
      (cond
       ((symbol? meta-filter) (begin
                                (code-append pig
                                             (assignment-stm
					      (get-ids-name new-var)
					      (cross-stm
                                               (par
                                                (distinct-stm
                                                 (foreach-stm
                                                  (get-meta-name to-be-filtered)
                                                  "$0")))
                                               (par
                                                (distinct-stm
                                                 (foreach-stm
                                                  (get-meta-name filter-dataset)
                                                  "$0"))))))
                                (code-append pig
                                             (assignment-stm
                                              (string-append (get-ids-name new-var) "_unmatched")
                                              (filter-stm
                                               (get-ids-name new-var)
                                               "$1 is null")))))
       ((string? meta-filter) (begin
                                (code-append pig
                                             (define-join-meta-filter new-var meta-filter))
                                (code-append pig
                                             (assignment-stm
                                              (get-ids-name new-var)
                                              (foreach-stm
                                               (par
                                                (filter-stm
                                                 (par
                                                  (foreach-stm
                                                   (par
                                                    (cross-stm
                                                     (par
                                                      (group-stm
                                                       (get-meta-name to-be-filtered)
                                                       "$0"))
                                                     (par
                                                      (group-stm
                                                       (get-meta-name filter-dataset)
                                                       "$0"))))
                                                   (comma-separated-list
                                                    "$0"
                                                    "$2"
                                                    (apply-function
                                                     (get-meta-filter-function-name new-var)
                                                     (par
                                                      (comma-separated-list
                                                       "$1" "$3"))))))
                                                 "$2==true"))
                                               (comma-separated-list "$0" "$1"))))
                                (code-append pig 
                                             (assignment-stm 
                                              (string-append (get-ids-name new-var) "_unmatched")
                                              (foreach-stm
                                               (par
                                                (filter-stm
                                                 (par
                                                  (string-append "JOIN "
                                                                 (par
                                                                  (distinct-stm
                                                                   (foreach-stm
                                                                    (get-meta-name to-be-filtered)
                                                                    "$0")))
                                                                 " by $0 left, "
                                                                 (get-ids-name new-var) " by $0"))
                                                 "$2 is  null"))
                                              (comma-separated-list "$0" "22061986L"))))))) ;i.e. a random number...just to fit the schema
      (code-append pig
		   (assignment-stm
		     (get-exp-group-name to-be-filtered)
		     (group-stm
		       (get-exp-name to-be-filtered)
		       "$0")))
      (code-append pig
		   (assignment-stm
		     (get-exp-group-name filter-dataset)
		     (group-stm
		       (get-exp-name filter-dataset)
		       "$0")))
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name new-var) "_partial")
		     (nested-foreach-stm
		       (par
			 (group-stm
			   (par
			     (foreach-stm
			       (par
				 (join-stm
				   (get-ids-name new-var) "$1"
				   (get-exp-group-name filter-dataset) "$0"))
			       (comma-separated-list
				 "$0" "flatten($3)")))
			   "$0"))
		       (list (assignment-stm
			       "cl"
			       (foreach-stm "$1" (comma-separated-list "$1" "$2" "$3"))))
		       (comma-separated-list
			 "$0" "cl"))))
      (code-append pig
                   (assignment-stm
                    (string-append (get-exp-name new-var) "_unmatched")
                    (foreach-stm
                     (par
                      (join-stm 
                       (string-append (get-ids-name new-var) "_unmatched")
                       "$0"
                       (get-exp-name to-be-filtered)
                       "$0"))
                     (comma-separated-list "$2" "$3" "$4"))))
      
      (code-append pig
		   (assignment-stm
		     (string-append (get-exp-name new-var) "_partial2")
		     (foreach-stm
		       (par
			 (foreach-stm
			   (par
			     (foreach-stm
			       (par
				 (join-stm
				   (get-exp-group-name to-be-filtered) "$0"
				   (string-append (get-exp-name new-var) "_partial") "$0"))
			       (apply-function
				 (get-difference-function new-var)
				 (apply-function "TOTUPLE" (comma-separated-list "$1" "$3")))))
			   "flatten($0)"))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "region")
			 (as-stm "$2" "value")))))
      (code-append pig 
                   (assignment-stm
                    (get-exp-name new-var) 
                    (foreach-stm
                     (par
                      (union-stm
                       (string-append (get-exp-name new-var) "_unmatched")
                       (string-append (get-exp-name new-var) "_partial2")))
                     (comma-separated-list
                      (as-stm "$0" "id")
                      (as-stm "$1" "region")
                      (as-stm "$2" "value")))))
      (code-append pig
		   (assignment-stm
		     (get-meta-group-name to-be-filtered)
		     (group-stm
		       (get-meta-name to-be-filtered)
		       "$0")))
      (code-append pig
		   (assignment-stm
		     (get-meta-group-name filter-dataset)
		     (group-stm
		       (get-meta-name filter-dataset)
		       "$0")))
      (code-append pig
		   (assignment-stm
		     (string-append (get-meta-name new-var) "_partial1")
		     (foreach-stm
		       (par
			 (join-stm
			   (get-ids-name new-var) "$0"
			   (get-meta-name to-be-filtered) "$0"))
		       (comma-separated-list
			 "$0"
			 (apply-function
			   "CONCAT"
			   (comma-separated-list
			     (string-append "'" (variable-identifier to-be-filtered) ".'")
			     "$3"))
			 "$4"))))
      (code-append pig
		   (assignment-stm
		     (string-append (get-meta-name new-var) "_partial2")
		     (foreach-stm
		       (par
			 (join-stm
			   (get-ids-name new-var) "$1"
			   (get-meta-name filter-dataset) "$0"))
		       (comma-separated-list
			 "$0"
			 (apply-function
			   "CONCAT"
			   (comma-separated-list
			     (string-append "'" (variable-identifier filter-dataset) ".'")
			     "$3"))
			 "$4"))))
      (code-append pig
                   (assignment-stm
                    (string-append (get-meta-name new-var) "_unmatched")
                    (foreach-stm
                    (par
                    (join-stm
                     (string-append (get-ids-name new-var) "_unmatched")
                     "$0"
                     (get-meta-name to-be-filtered)
                     "$0"))
                    (comma-separated-list "$2" "$3" "$4"))))
                                          
      (code-append pig
		   (assignment-stm
		     (get-meta-name new-var)
		     (foreach-stm
		       (par
                        (string-append "union "
			   (string-append (get-meta-name new-var) "_partial1") ", "
			   (string-append (get-meta-name new-var) "_partial2") ", "
                           (string-append (get-meta-name new-var) "_unmatched")))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "attribute")
			 (as-stm "$2" "value")))))
      (list preamble pig))))

;scm: schema
;function-list: list of function with parameters and output names
;  a element of the function-list can be:
;      - (list function new-attribute-name)
;      - (list function new-attribute-name parameter-name)
(define (emit-aggregate-code scm new-var function-list input-dataset)
  (letrec 
    ((input-schema-list (scm 'get (variable-identifier input-dataset) (variable-user input-dataset)))
     (input-schema-struct (cond
			    ((equal? (length input-schema-list) 0) (not-existing-schema-error input-dataset))
			    ((> (length input-schema-list) 1) (multiple-schema-definition-error scm input-dataset))
			    (else  (first input-schema-list))))
     (input-schema (second input-schema-struct))
     (input-name (variable-identifier input-dataset))
     (new-var-name (variable-identifier new-var))
     (right-name (variable-identifier input-dataset))
     (right-user (variable-user input-dataset))
     (pig "\n\n-- AGGREGATE \n")
     (preamble ""))
    (begin 
      (if
	(null? (scm 'get new-var-name 'any))
	(scm 'add new-var-name input-schema)
	(variable-redefinition-error new-var))
      (code-append pig
		   (assignment-stm
		     (get-exp-name new-var)
		     (foreach-stm
		       (get-exp-name input-dataset)
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "region")
			 (as-stm "$2" "value")))))
      (code-append pig
		   (assignment-stm
		     (string-append new-var-name "_" (get-exp-group-name input-dataset))
		     (group-stm
		       (get-exp-name input-dataset)
		       "$0")))
      (code-append pig
		   (assignment-stm
		     (get-meta-name new-var)
		     (foreach-stm
		       (par
			 (string-append
			   "UNION "
			   (get-meta-name input-dataset)
			   ","
			   (string-join
			     (map
			       (λ(f)
				 (string-append
				   (cond
				     ((equal? (first f) "COUNT") 
				      (let
					((path (if 
						 (list? (field-instance-path (second f)))
						 (field-instance-path (second f))
						 (list)))
					 (name (field-instance-identifier (second f))))
					(par
					  (foreach-stm
					    (string-append new-var-name "_" (get-exp-group-name input-dataset))
					    (comma-separated-list
					      "$0"
					      (string-append
						"'"
						(string-join (append 
							       (if 
								 (list? (field-instance-path (second f)))
								 (field-instance-path (second f))
								 (list))
							       (list (field-instance-identifier (second f))))
							     ".")
						"'")
					      (string-append 
						"(chararray)"
						(apply-function "COUNT" "$1")))))))
				     ((or (equal? (first f) "SUM") 
					  (equal? (first f) "AVG") 
					  (equal? (first f) "MIN")
					  (equal? (first f) "MAX")
                                          (equal? (first f) "BAG")
                                          (equal? (first f) "MEDIAN")
                                          (equal? (first f) "STD"))
				      (letrec
					((new-attribute-name (field-instance-identifier (second f)))
					 (new-attribute-path (if
							       (list? (field-instance-path (second f)))
							       (field-instance-path (second f))
							       (list)))
					 (function-name (first f))
					 (new-field-path (if
							   (list? (field-instance-path (third f)))
							   (field-instance-path (third f))
							   (list)))
					 (new-field-line (field-instance-line-number (third f)))
					 (new-field-name (field-instance-identifier (third f)))
					 (java-function-name
					   (cond
					     ((equal? (first f) "SUM") "GenometricPig.AggregateSum")
					     ((equal? (first f) "MAX") "GenometricPig.AggregateMax")
					     ((equal? (first f) "MIN") "GenometricPig.AggregateMin")
					     ((equal? (first f) "AVG") "GenometricPig.AggregateAvg")
                                             ((equal? (first f) "BAG") "GenometricPig.AggregateBag")
                                             ((equal? (first f) "MEDIAN") "GenometricPig.AggregateMedian")
                                             ((equal? (first f) "STD") "GenometricPig.AggregateStd")
                                             ))
					 (pos-list (scm 'position right-name right-user new-field-name new-field-path))
					 (new-field-position (cond
							       ((equal? (length pos-list) 1) (first pos-list))
							       ((equal? (length pos-list) 0) (not-existing-field-error
											       new-field-name
											       new-field-line
											       right-name))
							       (else (multiple-field-definition
								       new-field-name
								       new-field-line
								       right-name)))))
					(par
					  (foreach-stm
					    (par
					      (foreach-stm
						(string-append new-var-name "_" (get-exp-group-name input-dataset))
						(apply-function
						  java-function-name
						  (comma-separated-list "$0" 
									"$1" 
									(number->string new-field-position) 
									(string-append 
									  "'"
									  (string-join (append new-attribute-path (list new-attribute-name)) ".")
									  "'")))))
					    (comma-separated-list "$0.$0" "$0.$1" "$0.$2")))))
				     (else (error (string-append "Function " (first f) " has not been implemented yet."))))))
			       function-list)
			     ",")))
		       (comma-separated-list
			 (as-stm "$0" "id")
			 (as-stm "$1" "attribute")
			 (as-stm "$2" "value")))))
      (list preamble pig))))


;scm: schemata
;left: variable
;right: variable
;ordering: (list (ASC attname) (DESC attname) (...)..)
;top: (TOP N) (TOPG N)
(define (emit-order-code scm left right ordering top)
  (letrec
      ((pig "\n\n-- ORDER STATEMENT\n")
       (preamble "")
       (right-schema-list (scm 'get (variable-identifier right) (variable-user right)))
       (right-schema-struct (cond
                             ((equal? (length right-schema-list) 0) (not-existing-schema-error right))
                             ((> (length right-schema-list) 1) (multiple-schema-definition-error scm right))
                             (else  (first right-schema-list))))
       (right-schema (second right-schema-struct))
       (right-name (variable-identifier right))
       (right-user (variable-user right))
       (left-name (variable-identifier left)))
    (begin
      (if
       (null? (scm 'get left-name 'any))
       (scm 'add left-name right-schema)
       (variable-redefinition-error left)) 
      (code-append pig 
                   (define-order-function left ordering top)) 
      (code-append pig 
                   (assignment-stm
                    (get-meta-group-name left)
                    (group-all-stm
                    (par
                     (group-stm
                    (par 
                     (filter-stm
                     (get-meta-name right)
                     (string-join
                      (map 
                       (lambda (x)
                         (string-append "$1 == '"
                                        (string-join 
                                         (append (if (list? (field-instance-path (second x))) (field-instance-path (second x)) (list)) 
                                                 (list (field-instance-identifier (second x)))) ".")
                                        "'"))
                       ordering) " or ")))
                    "$0")))))
      (code-append pig 
                   (assignment-stm
                    (string-append (get-meta-name left) "_new")
                    (foreach-stm
                     (par
                      (foreach-stm
                       (get-meta-group-name left)
                       (apply-function (get-order-function left) "$1")))
                     (apply-function "flatten" "$0"))))
      (if
       (null? top)
       (begin
         (code-append pig 
                      (assignment-stm
                       (get-meta-name left)
                       (foreach-stm
                       (par
                       (union-stm
                        (get-meta-name right)
                        (string-append (get-meta-name left) "_new")))
                       (comma-separated-list
                        (as-stm "$0" "id")
                        (as-stm "$1" "region")
                        (as-stm "$2" "value")))))
         (code-append pig 
                      (assignment-stm
                       (get-exp-name left)
                       (get-exp-name right))))
       (begin 
         (code-append pig 
                      (assignment-stm
                       (get-meta-name-filter left)
                       (foreach-stm
                        (par
                         (join-stm
                          (get-meta-name right)
                          "$0"
                          (par
                           (foreach-stm
                            (string-append (get-meta-name left) "_new")
                            "$0"))
                          "$0"))
                        (comma-separated-list "$0" "$1" "$2")))) 
         (code-append pig 
                      (assignment-stm
                       (get-meta-name left)
                       (foreach-stm
                        (par
                         (union-stm
                          (get-meta-name-filter left)
                          (string-append (get-meta-name left) "_new")))
                        (comma-separated-list
                         (as-stm "$0" "id")
                         (as-stm "$1" "attribute")
                         (as-stm "$2" "value"))))) 
         (code-append pig 
                      (assignment-stm
                       (get-exp-name left)
                       (foreach-stm
                        (par
                         (join-stm
                          (get-exp-name right)
                          "$0"
                          (par
                           (foreach-stm
                            (string-append (get-meta-name left) "_new")
                            "$0"))
                          "$0"))
                        (comma-separated-list
                         "$0"
                         "$1"
                         "$2")))))) 
             
      (list preamble pig))))
