#lang racket

(require (prefix-in xml: "xml-utils.rkt"))
(require "configurations.rkt")
(require (prefix-in retro: "retro-utils.rkt"))

(provide 
  get-new-schemata
  is-indexed?
  get-type
  get-user
  merge-two-schemata
  append-field-to-schema
  schema-field
  schema-field-name
  schema-field-full-path
  schema-field-type
  compare-schemata)

;schemata structure is:
;(("name" "user" "type")
; (('type "name" "full-path")
;  ('type "name" "full-path")
;  ....
;  ))


(struct schema-header
	(name ;variable name			
	  type ;type: bed, narowPeak... or unknown
	  user ;owner of the dataset
	  lucene?)) ;is this dataset indexed?

(struct schema-field
	(type
	  name
	  [full-path #:mutable]))

;given an header, tells whether it is indexed or not
(define (is-indexed? h)
  (schema-header-lucene? h))

;given a schema struct (header, schema), returns the associated data type
(define (get-type s)
  (letrec
    ((header (first s))
     (type (schema-header-type header))
     (schema (second s)))
    (cond 
      ((equal? (string-downcase type) "gtf") (string-append
					       "gtf "
					       (if 
						 (> (length schema) 4)
						 (string-join (map (λ (f) (symbol->string (schema-field-type f))) (list-tail schema 4)) " ")
						 "")))
      ((equal? (string-downcase type) "tab") (string-append
					       "tab "
					       (string-join (map (λ (f) (symbol->string (schema-field-type f))) schema) " ")))
      ((equal? (string-downcase type) "tab1") (string-append
					       "tab1 "
					       (string-join (map (λ (f) (symbol->string (schema-field-type f))) schema) " ")))
      (else type))))

;given two schemata (list of fields)
;tells is they are equal or not
(define (compare-schemata scm1 scm2)
  (equal? scm1 scm2))

;given an headerm returns its owner
(define (get-user h)
  (schema-header-user h))

;utility functions for printing the schemata
(define (print-field fld)
  (let
    ((name (schema-field-name fld))
     (type (schema-field-type fld))
     (path (schema-field-full-path fld)))
    (displayln
      (string-append "\t"(foldr (λ (x y) (string-append x "." y)) name (if (list? path) path (list))) ":" (symbol->string type)))))
(define (print-variable var)
  (let
    ((header (first var))
     (field-list (second var)))
    (begin
      (displayln (string-append (schema-header-name header) ":"
				(schema-header-type header) ":" 
				(schema-header-user header) ":" 
				(format "~a" (schema-header-lucene? header))))
      (map
	print-field
	field-list)))) 

(define (print-schemata scm)
  (let loop ((lst scm))
    (when 
      (not (null? lst))
      (begin
	(print-variable (car lst))
	(loop (cdr lst))))))

;return the list of schema given a dataset and a owner
;if the owner is a string, it checks also for that, otherwise it is ignored
(define (get-schema scm name owner)
  (filter 
    (λ (x)
       (let
	 ((header (first x)))
	 (and
	   (equal? (schema-header-name header) name)
	   (if (string? owner) (equal? (schema-header-user header) owner) #t))))
    scm))


;return a xml string of the required schema
; this is used to print the output schema file
;in order to work this function requires that only
;one schema satisfies the query name+owner
(define (get-xml-schema scm name owner)
  (let
    ((schema (get-schema scm name owner)))
    (if
      (> (length schema) 1)
      (error "")
      (let
	((header (first (first schema)))
	 (field (second (first schema))))
	(string-append "<gmqlSchema name=\"" (schema-header-name header) "\" type=\""
		       (if (string? (schema-header-type header)) (schema-header-type header) "unknown") "\">\n"
		       (foldr 
			 (λ (x y) (string-append x "\n" y)) 
			 "</gmqlSchema>\n"
			 (map
			   (λ (x)
			      (string-append "<field type=\"" (symbol->string (schema-field-type x))
					     "\" > " 
					     (foldr (λ (x y) (string-append x "." y)) (schema-field-name x) (schema-field-full-path x) )
					     "</field>"))
			   field)))))))

;return the list of all the field positions in the dataset that
;satisfy the query
;Positions are zero based
(define (get-schema-position scm var owner name path)
  (let
    ((schema (cadar (get-schema scm var owner)))
     (result-list (list))) 
    (let loop ((pos 0) (lst schema))
      (if
	(null? lst)
	result-list
	(begin
	  (when (and 
		  (equal? (schema-field-name (car lst)) name)
		  (if (list? path) (equal? (schema-field-full-path (car lst)) path) #t))
	    (set! result-list (append result-list (list pos))))
	  (loop (+ 1 pos) (cdr lst)))))))

;get the list of the user that define a certain variable
(define (variable-name->user-list scm var-name)
  (map
    (λ (x) (schema-header-user (first x)))
    (filter
      (λ (x)
	 (let
	   ((header (first x)))
	   (equal? (schema-header-name header) var-name)))
      scm)))

;add a new variable to the schema list
(define (add-new-variable scm name schema)
  (letrec ((new-header (schema-header name "unknown" "runtime" #f))
	   (new-entry (list new-header schema)))
    (cons new-entry scm)))

;merge two schema and append the input variable names to the field path
;schemas: are schema-field lists
;var-names: are strings
(define (merge-two-schemata schema1 var-name1 schema2 var-name2)
  (let
    ((output-schema (list)))
    (append
      (map 
	(λ (x)
	   (let
	     ((name (schema-field-name x))
	      (path (cons var-name1 (schema-field-full-path x)))
	      (type (schema-field-type x)))
	     (schema-field type name path)))
	schema1)
      (map
	(λ (x)
	   (let
	     ((name (schema-field-name x))
	      (path (cons var-name2 (schema-field-full-path x)))
	      (type (schema-field-type x)))
	     (schema-field type name path)))
	schema2)
      )))

;append a field at the end of the given schema
;the path of the appended field will be empty
(define (append-field-to-schema schema field-name field-type)
  (append schema (list (schema-field field-type field-name (list)))))


;get the list of all the variable currentlyin the schema collection
;returns (list (list name user) ....)
(define (get-list-defined-variable scm)
  (map
    (λ (x) (let 
	      ((header (first x)))
	      (list (schema-header-name header) (schema-header-user header))))
    scm))

;parses the data-schemata XML tree and produces
; a list of all the schemata
; the user argument is the identifier of the user that owns the dataset
;name is the dataset name::string (from the path)
(define (xml->var x user name)
  (match x
	 [(list 'gqlSchemaCollection attributes r ...) (map (λ (x) (xml->var x user name)) (xml:trim-list r))]
         [(list 'gmqlSchemaCollection attributes r ...) (map (λ (x) (xml->var x user name)) (xml:trim-list r))]
	 [(list 'gqlSchema attributes r ...) (list (schema-header
						     name
						     (retro:string-trim (xml:get-optional-attribute-value 'type "unknown" attributes))
						     user
						     #t)
						   (filter (lambda (x) (not (equal? (schema-field-type x) 'STRAND))) (map (λ (x) (xml->var x user name)) (xml:trim-list r))))]
	 [(list 'gmqlSchema attributes r ...) (list (schema-header
						     name
						     (retro:string-trim (xml:get-optional-attribute-value 'type "unknown" attributes))
						     user
						     #t)
						   (filter (lambda (x) (not (equal? (schema-field-type x) 'STRAND))) (map (λ (x) (xml->var x user name)) (xml:trim-list r))))]
         [(list 'field attributes name) (schema-field 
					  (string->symbol (retro:string-trim (xml:get-attribute-value 'type attributes)))
					  (retro:string-trim name)
					  (list))]))


;given a path in <path: var_name.schema> returns a string "var-name"
(define (schema-path->var-name path)
  (let
    ((path-string (path->string path)))
    (substring path-string 0 (- (string-length path-string) 7))))  ;wher 7 is the lenght of the ".schema" extension

;given the list of user, fill the schemata
(define (fill-schemata user-list)
  (let
    ((new-schemata (list)))
    (begin
      (map
	(λ (username) 
	   (letrec
	     ((schema-dir (string-append repository-root "/" username "/schema/" ))
	      (schema-list (directory-list schema-dir)))
	     (map 
	       (λ (file-name)
		  (letrec
		    ((file-url (string-append schema-dir (path->string file-name)))
		     (schema-xexpr (xml:file->xexpr file-url)))
                    (set! new-schemata (append (xml->var schema-xexpr username (schema-path->var-name file-name)) new-schemata))))
	       schema-list)))
	user-list)
      new-schemata)))

;main class for handling the schemata repository
; it returns a closure
(define (get-new-schemata)
  (let
    ((schemata '()))
    (λ args
       (if
	 (equal? args '())
	 (error "the schemata closure need at least one argument")
	 (let
	   ((cmd (first args)))
	   (cond
	     [(not (symbol? cmd)) (error "First argument for the schemata closure has to be a symbol")]
	     [(equal? cmd 'print) (print-schemata schemata)]
	     [(equal? cmd 'fill)
	      (if 
		(null? (cdr args))
		(error "user list is empty, but it shouldn't")
		(let
		  ((user-list (cadr args)))
		  (set! schemata (fill-schemata user-list))))]
	     [(equal? cmd 'get)
	      (let ((name (second args))
		    (user (third args)))
		(get-schema schemata name user))]
	     [(equal? cmd 'get-xml)
	      (let ((name (second args))
		    (user (third args)))
		(get-xml-schema schemata name user))]
	     [(equal? cmd 'variable-name->user-list)
	      (let ((name (second args)))
		(variable-name->user-list schemata name))]
	     [(equal? cmd 'add)
	      (let ((name (second args))
		    (schema (third args)))
		(set! schemata (add-new-variable schemata name schema)))]
	     [(equal? cmd 'get-list)  (get-list-defined-variable schemata)]
	     [(equal? cmd 'position) 
	      (let ((var-name (second args))
		    (var-owner (third args))
		    (field-name (fourth args))
		    (field-path (fifth args)))
		(get-schema-position schemata var-name var-owner field-name field-path))]
	     [else (error (string-append "Argument '" (symbol->string cmd) "' is not valid for the schemata closure"))]))))))
